%
%
% This example runs an algorithm that is a mixture of REPS and Pi2. 
% The path to be optimized comes from the mean of several "a" letters.
%
% The mean is obtained as a traditional ProMP but the ProMP itself is not used for the
% optimization. While you could use the covariance matrix of the ProMP to generate exploration
% noise, for this example, I think the STOMP-like covariance matrix based on finite differencing
% works much better. The reason is that the problem is basically a motion planning problem that has
% to avoid an obstacle and, at the same time, reach a via-point. The path has to be largely
% deformed. If you just want to do small fine-tunning of a demonstrated trajectory, then the
% original covariance of the ProMP could work better.
%
% After the optimization is finished, I reuse the ProMP covariance just to illustrate how to
% maintain the ProMP distribution.
%
% The main parts of the code that will need tunning and are problem dependent are the definition of
% the cost function (getCostLetterA.m) and their weights.
%
% The weights are adjusted in the file userParamCostREPS.m
% Also, you may need to adjust the number of roll-outs and updates in the file userParamREPS.m
%
% Some points of interest:
%
%   1. Large number of roll-outs (param.reps.n_rollOouts) generate better estimates of the update.
%      Ideally, one would like to have a very large number. This number is basically limited by how
%      expensive it is to get one roll-out.
%
%   2. The shape and magnitude  of the covariance matrix to generate exploration noise (sys(ii).dof(j).wCov) is
%      decisive to achieve a suitable exploration, and therefore, faster convergence. For motion
%      planning problems, the STOMP-like covariance seems to work well. However, the STOMP type is not suited
%      when you need to change the initial and final goal because it does not perturb the
%      extremities of the path. In this case I added a hack where the initial and final weights can
%      be perturbed, but it is just a hack. The ideal solution would be to design a proper
%      covariance matrix.
%
%   


clear; clc; close all; dbstop if error;
initialize;


%% Load data and create ProMP

% load the letter "a" to create a 2DoF problem
dt     = 0.01;
[data] = loadData( dt );

% put data in the ClassSystem-like format
data_ias = format_ias_data(data);

% create ProMP here
ProMP = regression_pmp(data_ias, 30, 'GaussianFeatures');
ProMP.dt = dt;

% plot the initial distribution
for j=1:2
    ir = ProMP.w.index{j};
    Gn = ProMP.basis.Gn;
    Gndot = ProMP.basis.Gndot;
    Sigma_model = ProMP.w.cov_full(ir, ir);
    Sigma_q     = Gn*Sigma_model*Gn';
    mean_q      = Gn*ProMP.w.mean_full(ir);
    mean_qdot   = Gndot*ProMP.w.mean_full(ir);
    h.traj(j).fig = figurew(['dof' num2str(j)]);
    if j==1  
        set_fig_position([0.104 0.569 0.305 0.408]);
    else
        set_fig_position([0.103 0.147 0.305 0.408]);
    end
    hShade = shadedErrorBar([1:ProMP.nTraj], mean_q , 2*sqrt( diag(Sigma_q) ), {'g', 'LineWidth', 2'}, 1);
    delete(hShade.mainLine); delete(hShade.edge);
    givenTraj.q(:,j)    = mean_q;
    givenTraj.qdot(:,j) = mean_qdot;
end

%% Specify via point

param.dof(1).viaPoint = [80   0.7];
param.dof(2).viaPoint = [80   0.7];
for j=1:2
    figure(h.traj(j).fig)
    plot(param.dof(j).viaPoint(:,1), param.dof(j).viaPoint(:,2), sty('b', 'o', 0.1, 'none', 10) );
end

%% Create and plot obstacle
obs = createObstacle();
h.traj(3).fig = figurew('XY final trajectory'); axis 'equal';
xlabel('x')
ylabel('y');
title('2D path of point mass');
plot(givenTraj.q(:,1), givenTraj.q(:,2),  sty('b', [], 3) );
plot(param.dof(1).viaPoint(2), param.dof(2).viaPoint(2), sty('b', 'o', 0.1, 'none', 10));
obs = plotObstacle(obs);


%% Specify REPS parameters
param.reps     = userParamREPS();
param.repsCost = userParamCostREPS( givenTraj.q );
param.obstacle = obs;


%% Prepare the main sys structure. 
%  Each dof is uncorrelated from each other: we assume that each dof trajectory will be optimized
%  independent from each other.
ii=1;
sys(ii).name     = 'trajectory';
sys(ii).Cov      =  ProMP.w.cov_full;
sys(ii).basis    =  ProMP.basis;
sys(ii).nBasis   =  ProMP.nBasis;
for j=1:2 % dofs of interest are x and y
    % uncorrelate dofs
    sys(ii).dof(j).dt     = ProMP.dt;
    sys(ii).dof(j).q      = givenTraj.q(:,j);
    sys(ii).dof(j).qdot   = givenTraj.qdot(:,j);
    sys(ii).dof(j).nBasis = ProMP.nBasis;
    sys(ii).dof(j).wMean  = ProMP.w.mean_full(ProMP.w.index{j});
    sys(ii).dof(j).wCov(:,:,1) = ProMP.w.cov_full(ProMP.w.index{j}, ProMP.w.index{j});
end                                          
sys(ii).nDoF = numel(sys(ii).dof);


%% Run 1: Lock the initial and end of the patch


% Modify the covariance matrix of each joint to explore as in STOMP (see STOMP ICRA2011 paper)
if param.reps.useSTOMPCovariance
    for j=1:2
        sys(ii).dof(j).wCov(:,:,1) = param.reps.covarInitialScale*covarianceStomp( ProMP.nBasis );
    end
end


% Main optimization and plot of result.
% Note that the ProMP itself is not required. Only the parameterization of the mean trajectory is
% actually important.
sys2 = main_loop(param, sys, h);


% plot result with the original distribution but with the new optimized mean
for j=1:2    
    wMean = sys2.dof(j).wMean(:,end);
    wCov  = ProMP.w.cov_full(ProMP.w.index{j}, ProMP.w.index{j}); % use the original covariance
    Gn = sys2.basis.Gn;
    mean_q  = Gn*wMean;
    Sigma_q = Gn*wCov*Gn';    
    figure(h.traj(j).fig);
    hShade = shadedErrorBar([1:ProMP.nTraj], mean_q , 2*sqrt( diag(Sigma_q) ), {'b', 'LineWidth', 2'}, 1);
    delete(hShade.mainLine); delete(hShade.edge);
end

pause

%% Run 2: Let the extremities be optimized as well.

fprintf('\n\n*********************************\n');
fprintf('Running again, but allowing the initial and end point to change\n');
fprintf('*********************************\n');

obs = createObstacle();
h.traj(3).fig = figurew('XY final trajectory 2'); axis 'equal';
xlabel('x')
ylabel('y');
title('2D path of point mass');
plot(givenTraj.q(:,1), givenTraj.q(:,2),  sty('b', [], 3) );
plot(param.dof(1).viaPoint(2), param.dof(2).viaPoint(2), sty('b', 'o', 0.1, 'none', 10));
obs = plotObstacle(obs);

% modify initial parameters to make the optimized trajectory "detach" from the initial mean.
param.reps.anchorInitial_nWeights = 0;
param.reps.anchorFinal_nWeights = 0;
param.reps.shakeInitial_nWeights = 3;
param.reps.shakeFinal_nWeights = 3;
param.repsCost.deviationFromMean = 0;

sys3 = main_loop(param, sys, h);


% plot result with the original distribution but with the new optimized mean
for j=1:2    
    wMean = sys3.dof(j).wMean(:,end);
    wCov  = ProMP.w.cov_full(ProMP.w.index{j}, ProMP.w.index{j}); % use the original covariance
    Gn = sys3.basis.Gn;
    mean_q  = Gn*wMean;
    Sigma_q = Gn*wCov*Gn';    
    figure(h.traj(j).fig);
    hShade = shadedErrorBar([1:ProMP.nTraj], mean_q , 2*sqrt( diag(Sigma_q) ), {'r', 'LineWidth', 2'}, 1);
    delete(hShade.mainLine); delete(hShade.edge);
end
















